"""Size limit global variables."""


MAX_ROWS = 2000  # Maximum number of sents/rows
MAX_SENT_LENGTH = 100  # Maximum words in each sentence
MAX_WORD_LENGTH = 100  # Maximum length of each word

MAX_CATEGORIES = 20  # Maximum number of categories in a categorical field.
